import numpy as np
from numpy.linalg import solve

def hatfun(t, k):
    """
    hatfun(t, k)

    Returns a piecewise linear "hat" function,  where t is a vector of
    n+1 interpolation nodes and k is an integer in 0:n giving the index of the node
    where the hat function equals one.
    """
    n = len(t) - 1

    def evaluate(x):
        H = np.zeros(np.array(x).shape)
        for (j, xj) in enumerate(x):
            if (k > 0) and (t[k-1] <= xj) and (xj <= t[k]):
                H[j] = (xj - t[k-1]) / (t[k] - t[k-1])
            elif (k < n) and (t[k] <= xj) and (xj <= t[k+1]):
                H[j] = (t[k+1] - xj) / (t[k+1] - t[k])
        return H
    return evaluate

def plinterp(t, y):
    """
    plinterp(t, y)

    Create a piecewise linear interpolating function for data values in y given at nodes
    in t.
    """
    n = len(t) - 1
    H = [hatfun(t, k) for k in range(n+1)]
    def evaluate(x):
        f = 0
        for k in range(n+1):
            f += y[k] * H[k](x)
        return f
    return evaluate

def spinterp(t, y):
    """
    spinterp(t, y)

    Create a cubic not-a-knot spline interpolating function for data values in y given at nodes in t.
    """
    n = len(t) - 1
    h = [t[i + 1] - t[i] for i in range(n)]

    # Preliminary definitions.
    Z = np.zeros([n, n])
    I = np.eye(n)
    E = I[: n - 1, :]
    J = np.eye(n) + np.diag(-np.ones(n - 1), 1)
    H = np.diag(h)

    # Left endpoint interpolation:
    AL = np.hstack([I, Z, Z, Z])
    vL = y[:-1]

    # Right endpoint interpolation:
    AR = np.hstack([I, H, H**2, H**3])
    vR = y[1:]

    # Continuity of first derivative:
    A1 = E @ np.hstack([Z, J, 2 * H, 3 * H**2])
    v1 = np.zeros(n - 1)

    # Continuity of second derivative:
    A2 = E @ np.hstack([Z, Z, J, 3 * H])
    v2 = np.zeros(n - 1)

    # Not-a-knot conditions:
    nakL = np.hstack([np.zeros(3 * n), np.hstack([1, -1, np.zeros(n - 2)])])
    nakR = np.hstack([np.zeros(3 * n), np.hstack([np.zeros(n - 2), 1, -1])])

    # Assemble and solve the full system.
    A = np.vstack([AL, AR, A1, A2, nakL, nakR])
    v = np.hstack([vL, vR, v1, v2, 0, 0])
    z = solve(A, v)

    # Break the coefficients into separate vectors.
    rows = np.arange(n)
    a = z[rows]
    b = z[n + rows]
    c = z[2 * n + rows]
    d = z[3 * n + rows]
    S = [np.poly1d([d[k], c[k], b[k], a[k]]) for k in range(n)]

    # This function evaluates the spline when called with a value for x.
    def evaluate(x):
        f = np.zeros(x.shape)
        for k in range(n):
            # Evaluate this piece's cubic at the points inside it.
            index = (x >= t[k]) & (x <= t[k + 1])
            f[index] = S[k](x[index] - t[k])
        return f

    return evaluate

def fdweights(t, m):
    """
    fdweights(t, m)

    Return weights for the mth derivative of a function at zero using values at the
    nodes in vector t.
    """
    # This is a compact implementation, not an efficient one.

    def weight(t, m, r, k):
        # Recursion for one weight.
        # Input:
        #   t   nodes (vector)
        #   m   order of derivative sought
        #   r   number of nodes to use from t (<= length(t))
        #   k   index of node whose weight is found

        if (m < 0) or (m > r):  # undefined coeffs must be zero
            c = 0
        elif (m == 0) and (r == 0):  # base case of one-point interpolation
            c = 1
        else:  # generic recursion
            if k < r:
                denom = t[r] - t[k]
                c = (t[r] * weight(t, m, r-1, k) - m * weight(t, m-1, r-1, k)) / denom
            else:
                beta = np.prod(t[r-1] - t[:r-1]) / np.prod(t[r] - t[:r])
                c = beta * (m * weight(t, m-1, r-1, r-1) - t[r-1] * weight(t, m, r-1, r-1))
        return c

    r = len(t) - 1
    w = np.zeros(t.shape)
    return np.array([ weight(t, m, r, k) for k in range(r+1) ])

def trapezoid(f, a, b, n):
    """
    trapezoid(f, a, b, n)

    Apply the trapezoid integration formula for integrand f over interval [a,b], broken up into n equal pieces. Returns estimate, vector of nodes, and vector of integrand values at the nodes.
    """
    h = (b - a) / n
    t = np.linspace(a, b, n + 1)
    y = f(t)
    T = h * (np.sum(y[1:-1]) + 0.5 * (y[0] + y[-1]))
    return T, t, y

def intadapt(f, a, b, tol):
    """
    intadapt(f, a, b, tol)

    Do adaptive integration to estimate the integral of f over [a,b] to desired
    error tolerance tol. Returns estimate and a vector of evaluation nodes used.
    """

    # Use error estimation and recursive bisection.
    def do_integral(a, fa, b, fb, m, fm, tol):
        # These are the two new nodes and their f-values.
        xl = (a + m) / 2
        fl = f(xl)
        xr = (m + b) / 2
        fr = f(xr)
        t = np.array([a, xl, m, xr, b])  # all 5 nodes at this level

        # Compute the trapezoid values iteratively.
        h = b - a
        T = np.zeros(3)
        T[0] = h * (fa + fb) / 2
        T[1] = T[0] / 2 + (h / 2) * fm
        T[2] = T[1] / 2 + (h / 4) * (fl + fr)

        S = (4 * T[1:] - T[:-1]) / 3  # Simpson values
        E = (S[1] - S[0]) / 15  # error estimate

        if abs(E) < tol * (1 + abs(S[1])):  # acceptable error?
            Q = S[1]  # yes--done
        else:
            # Error is too large--bisect and recurse.
            QL, tL = do_integral(a, fa, m, fm, xl, fl, tol)
            QR, tR = do_integral(m, fm, b, fb, xr, fr, tol)
            Q = QL + QR
            t = np.hstack([tL, tR[1:]])  # merge the nodes w/o duplicate
        return Q, t

    m = (b + a) / 2
    Q, t = do_integral(a, f(a), b, f(b), m, f(m), tol)
    return Q, t



import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np
from numpy.linalg import norm

V = np.random.randn(4,4)
D = np.diag([-2,0.4,-0.1,0.01])
A = V @ D @ np.linalg.inv(V)
b = np.ones(4)

def test_poweriter():
    beta,x = FNC.poweriter(A,30)
    assert isclose(beta[-1], -2, rtol=1e-10)
    d = np.dot(x,V[:,0])/(norm(V[:,0])*norm(x))
    assert isclose(np.abs(d), 1, rtol=1e-10)

def test_inviter():
    beta,x = FNC.inviter(A,0.37,15)
    assert isclose(beta[-1], 0.4, rtol=1e-10)
    d = np.dot(x,V[:, 1]) / (norm(V[:, 1])*norm(x))
    assert isclose(np.abs(d), 1, rtol=1e-10)

def test_arnoldi():
    Q,H = FNC.arnoldi(A, b, 4)
    assert isclose(A @ Q[:, :4], Q @ H, rtol=1e-10).all()

def test_gmres():
    x,res = FNC.gmres(A, b, 3)
    assert isclose(norm(b - A @ x), res[-1], rtol=1e-10)
    x,res = FNC.gmres(A, b, 4)
    assert isclose(A @ x, b, rtol=1e-10).all()



from scipy.integrate import solve_ivp
import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np

f = lambda t, u: u - 2*t**2
u_ex = np.exp(1.5) - 2*(-2 + 2*np.exp(1.5) - 2*1.5 - 1.5**2)
g = lambda t, u: np.array([t -6 - np.sin(u[1]), u[0]])
sol = solve_ivp(g, [1, 2], np.array([-1., 4]), rtol=1e-11, atol=1e-11)

def test_euler():
    t, u = FNC.euler(f, [0, 1.5], 1.0, 4000)
    assert isclose(u[0,-1], u_ex, rtol=0.005)

def test_ie2():
    t, u = FNC.ie2(f, [0, 1.5], 1.0, 4000)
    assert isclose(u[:, -1], u_ex, rtol=0.0001)

def test_am2():
    t, u = FNC.am2(f, [0, 1.5], 1.0, 4000)
    assert isclose(u[:, -1], u_ex, rtol=1e-4)

def test_eulersys():
    t, u = FNC.euler(g, [1, 2], [-1., 4], 4000)
    assert isclose(u[:, -1], sol.y[:, -1], rtol=0.004).all()

def test_ie2sys():
    t, u = FNC.ie2(g, [1, 2], [-1., 4], 4000)
    assert isclose(u[:, -1], sol.y[:, -1], rtol=1e-3).all()

def test_rk4():
    t, u = FNC.rk4(g, [1, 2], [-1., 4], 800)
    assert isclose(u[:, -1], sol.y[:,-1], rtol=1e-6).all()

def test_rk23():
    t, u = FNC.rk23(g, [1, 2], [-1., 4], 1e-5)
    assert isclose(u[:, -1], sol.y[:, -1], rtol=1e-4).all()

def test_ab4():
    t, u = FNC.ab4(g, [1, 2], [-1., 4], 800)
    assert isclose(u[:, -1], sol.y[:, -1], rtol=1e-4).all()

def test_am2sys():
    t, u = FNC.am2(g, [1, 2], [-1., 4], 2000)
    assert isclose(u[:, -1], sol.y[:, -1], rtol=1e-4).all()



import numpy as np
from scipy.integrate import solve_ivp
from .chapter04 import levenberg

def shoot(phi, a, b, ga, gb, init):
    """
    shoot(phi, a, b, ga, gb, init)

    Use the shooting method to solve a two-point boundary value problem. 
    The ODE is u'' = phi(x, u, u') for x in (a,b). The functions 
    ga(u(a), u'(a)) and gb(u(b), u'(b)) specify the boundary conditions. 
    The value init is an initial guess for [u(a), u'(a)].

    Return vectors for the nodes, the values of u, and the values of u'.
    """

    # Tolerances for IVP solver and rootfinder.
    tol = 1e-5
    # To be solved by the IVP solver
    def shootivp(x, y):
        return np.array([y[1], phi(x, y[0], y[1])])

    # Evaluate the difference between computed and target values at x=b.
    def objective(s):
        nonlocal x, y  # change these values in outer scope

        x = np.linspace(a, b, 400)  # make decent plots on return
        sol = solve_ivp(shootivp, [a, b], s, atol=tol/10, rtol=tol/10, t_eval=x)
        x = sol.t
        y = sol.y
        residual = np.array([ga(y[0, 0], y[1, 0]), gb(y[0, -1], y[0, -1])])
        return residual

    # Find the unknown quantity at x=a by rootfinding.
    x, y = [], []    # the values will be overwritten
    s = levenberg(objective, init, tol)

    # Don't need to solve the IVP again. It was done within the
    # objective function already.
    u = y[0]        # solution
    du_dx = y[1]    # derivative

    return x, u, du_dx


def diffmat2(n, xspan):
    """
    diffmat2(n, xspan)

    Compute 2nd-order-accurate differentiation matrices on n+1 points in the
    interval xspan. Return a vector of nodes, and the matrices for the first
    and second derivatives.
    """
    a, b = xspan
    h = (b - a) / n
    x = np.linspace(a, b, n + 1)  # nodes

    # Define most of Dx by its diagonals.
    dp = 0.5 / h * np.ones(n)  # superdiagonal
    dm = -0.5 / h * np.ones(n)  # subdiagonal
    Dx = np.diag(dm, -1) + np.diag(dp, 1)

    # Fix first and last rows.
    Dx[0, :3] = np.array([-1.5, 2, -0.5]) / h
    Dx[-1, -3:] = np.array([0.5, -2, 1.5]) / h

    # Define most of Dxx by its diagonals.
    d0 = -2 / h**2 * np.ones(n + 1)  # main diagonal
    dp = np.ones(n) / h**2  # superdiagonal and subdiagonal
    Dxx = np.diag(d0, 0) + np.diag(dp, -1) + np.diag(dp, 1)

    # Fix first and last rows.
    Dxx[0, :4] = np.array([2, -5, 4, -1]) / h**2
    Dxx[-1, -4:] = np.array([-1, 4, -5, 2]) / h**2

    return x, Dx, Dxx


def diffcheb(n, xspan):
    """
    diffcheb(n, xspan)

    Compute Chebyshev differentiation matrices on n+1 points in the
    interval xspan. Return a vector of nodes, and the matrices for the first
    and second derivatives.
    """
    x = -np.cos(np.arange(n + 1) * np.pi / n)  # nodes in [-1,1]
    Dx = np.zeros([n + 1, n + 1])
    c = np.hstack([2.0, np.ones(n - 1), 2.0])  # endpoint factors

    # Off-diagonal entries
    Dx = np.zeros([n + 1, n + 1])
    for i in range(n + 1):
        for j in range(n + 1):
            if i != j:
                Dx[i, j] = (-1) ** (i + j) * c[i] / (c[j] * (x[i] - x[j]))

    # Diagonal entries by the "negative sum trick"
    for i in range(n + 1):
        Dx[i, i] = -np.sum([Dx[i, j] for j in range(n + 1) if j != i])

    # Transplant to [a,b]
    a, b = xspan
    x = a + (b - a) * (x + 1) / 2
    Dx = 2 * Dx / (b - a)

    # Second derivative
    Dxx = Dx @ Dx

    return x, Dx, Dxx


def bvplin(p, q, r, xspan, lval, rval, n):
    """
        bvplin(p, q, r, xspan, lval, rval, n)

    Use finite differences to solve a linear bopundary value problem. The ODE is
    u''+p(x)u'+q(x)u = r(x) on the interval xspan, with endpoint function
    values given as lval and rval. There will be n+1 equally spaced nodes,
    including the endpoints.

    Return vectors of the nodes and the solution values.
    """
    x, Dx, Dxx = diffmat2(n, xspan)

    P = np.diag(p(x))
    Q = np.diag(q(x))
    L = Dxx + P @ Dx + Q  # ODE expressed at the nodes

    # Replace first and last rows using boundary conditions.
    I = np.eye(n + 1)
    A = np.vstack([I[0], L[1:-1], I[-1]])
    b = np.hstack([lval, r(x[1:-1]), rval])

    # Solve the system.
    u = np.linalg.solve(A, b)

    return x, u

def bvp(phi, xspan, ga, gb, init):
    """
    bvp(phi, xspan, ga, gb, init)

    Use finite differences to solve a two-point boundary value problem. 
    The ODE is u'' = phi(x, u, u') for x in (a,b). The functions 
    ga(u(a), u'(a)) and gb(u(b), u'(b)) specify the boundary conditions. 
    The value init is an initial guess for [u(a), u'(a)].

    Return vectors for the nodes and the values of u.
    """
    n = len(init) - 1
    x, Dx, Dxx = diffmat2(n, xspan)
    h = x[1] - x[0]
    def residual(u):
        # Compute the difference between u'' and phi(x,u,u') at the
        # interior nodes and appends the error at the boundaries.
        du_dx = Dx @ u  # discrete u'
        d2u_dx2 = Dxx @ u  # discrete u''
        f = d2u_dx2 - phi(x, u, du_dx)

        # Replace first and last values by boundary conditions.
        f[0] = ga(u[0], du_dx[0]) / h
        f[n] = gb(u[n], du_dx[n]) / h
        return f

    u = levenberg(residual, init.copy())
    return x, u[-1]


def fem(c, s, f, a, b, n):
    """
    fem(c, s, f, a, b, n)

    Use a piecewise linear finite element method to solve a two-point boundary
    value problem. The ODE is (c(x)u')' + s(x)u = f(x) on the interval
    [a,b], and the boundary values are zero. The discretization uses n equal
    subintervals.

    Return vectors for the nodes and the values of u.
    """
    # Define the grid.
    h = (b - a) / n
    x = np.linspace(a, b, n + 1)

    # Templates for the subinterval matrix and vector contributions.
    Ke = np.array([[1, -1], [-1, 1]])
    Me = (1 / 6) * np.array([[2, 1], [1, 2]])
    fe = (1 / 2) * np.array([1, 1])

    # Evaluate coefficent functions and find average values.
    cval = c(x)
    cbar = (cval[:-1] + cval[1:]) / 2
    sval = s(x)
    sbar = (sval[:-1] + sval[1:]) / 2
    fval = f(x)
    fbar = (fval[:-1] + fval[1:]) / 2

    # Assemble global system, one interval at a time.
    K = np.zeros([n - 1, n - 1])
    M = np.zeros([n - 1, n - 1])
    f = np.zeros(n - 1)
    K[0, 0] = cbar[0] / h
    M[0, 0] = sbar[0] * h / 3
    f[0] = fbar[0] * h / 2
    K[-1, -1] = cbar[-1] / h
    M[-1, -1] = sbar[-1] * h / 3
    f[-1] = fbar[-1] * h / 2
    for k in range(1, n - 1):
        K[k - 1 : k + 1, k - 1 : k + 1] += (cbar[k] / h) * Ke
        M[k - 1 : k + 1, k - 1 : k + 1] += (sbar[k] * h) * Me
        f[k - 1 : k + 1] += (fbar[k] * h) * fe

    # Solve system for the interior values.
    u = np.linalg.solve(K + M, f)
    u = np.hstack([0, u, 0])  # put the boundary values into the result

    return x, u



import numpy as np
from .chapter04 import levenberg
from .chapter10 import diffcheb
from scipy.integrate import solve_ivp

def diffper(n, xspan):
    """
    diffper(n, xspan)

    Construct 2nd-order differentiation matrices for functions with periodic end
    conditions, using `n` unique nodes in the interval `xspan`. Return a vector of
    nodes and the  matrices for the first and second derivatives.
    """
    a, b = xspan
    h = (b - a) / n
    x = a + h * np.arange(n)  # nodes, omitting the repeated data

    # Construct Dx by diagonals, then correct the corners.
    dp = 0.5 / h * np.ones(n - 1)  # superdiagonal
    dm = -0.5 / h * np.ones(n - 1)  # subdiagonal
    Dx = np.diag(dm, -1) + np.diag(dp, 1)
    Dx[0, -1] = -1 / (2 * h)
    Dx[-1, 0] = 1 / (2 * h)

    # Construct Dxx by diagonals, then correct the corners.
    d0 = -2 / h**2 * np.ones(n)  # main diagonal
    dp = np.ones(n - 1) / h**2  # superdiagonal and subdiagonal
    Dxx = np.diag(d0) + np.diag(dp, -1) + np.diag(dp, 1)
    Dxx[0, -1] = 1 / (h**2)
    Dxx[-1, 0] = 1 / (h**2)

    return x, Dx, Dxx


def parabolic(phi, xspan, m, ga, gb, tspan, init):
    """
        parabolic(phi, xspan, m, ga, gb, tspan, init)

    Solve a parabolic PDE by the method of lines. The PDE is 
    ∂u/∂t = phi(t,x,u,∂u/∂x,∂^2u/∂x^2), xspan gives the space 
    domain, m gives the degree of a Chebyshev spectral discretization, ga and gb are functions of (u,∂u/∂x) at the domain ends that should be made zero, tspan is the time domain, and init is a function of x that gives the initial condition. Returns a vector x and a function of t that gives the semidiscrete solution at x. 
    """   
    x, Dx, Dxx = diffcheb(m, xspan)
    int = range(1, m)    # indexes of interior nodes

    def extend(v):
        def objective(ubc):
            u0, um = ubc
            ux = Dx @ np.hstack([u0, v, um])
            return np.array([ga(u0, ux[1]), gb(um, ux[-1])])
        ubc = levenberg(objective, np.array([0, 0]))[-1]
        return np.hstack([ubc[0], v, ubc[-1]])

    def ode(t, v):
        u = extend(v)
        ux, uxx = Dx @ u, Dxx @ u
        return phi(t, x[int], u[int], ux[int], uxx[int])

    u0 = init(x[int])
    sol = solve_ivp(ode, tspan, u0, method="BDF", dense_output=True)

    return x, lambda t: extend(sol.sol(t))



import numpy as np
from .chapter04 import levenberg
import warnings

def euler(du_dt, tspan, u0, n):
    """
    euler(du_dt, tspan, u0, n)

    Apply Euler's method to solve the IVP u'=du_dt(u,t) over the interval tspan with
    u(tspan[1])=u0, using n subintervals/steps. Return vectors of times and solution
    values.
    """
    a, b = tspan
    h = (b - a) / n
    t = np.linspace(a, b, n+1)
    u = np.tile(np.array(u0), (n+1, 1))
    for i in range(n):
        u[i+1] = u[i] + h * du_dt(t[i], u[i])

    return t, u.T

def ie2(du_dt, tspan, u0, n):
    """
    ie2(du_dt, tspan, u0, n)

    Apply the Improved Euler method to solve the vector-valued IVP u'=du_dt(u,p,t) over the
    interval tspan with u(tspan[1])=u0, using n subintervals/steps. Returns a vector
    of times and a vector of solution values/vectors.
    """
    # Time discretization.
    a, b = tspan
    h = (b - a) / n
    t = np.linspace(a, b, n + 1)

    # Initialize output.
    u = np.tile(np.array(u0), (n+1, 1))

    # Time stepping.
    for i in range(n):
        uhalf = u[i] + h / 2 * du_dt(t[i], u[i])
        u[i+1] = u[i] + h * du_dt(t[i] + h / 2, uhalf)

    return t, u.T

def rk4(du_dt, tspan, u0, n):
    """
    rk4(du_dt, tspan, u0, n)

    Apply "the" Runge-Kutta 4th order method to solve the vector-valued IVP u'=du_dt(u,p,t)
    over the interval tspan with u(tspan[1])=u0, using n subintervals/steps.
    Return a vector of times and a vector of solution values/vectors.
    """
    # Time discretization.
    a, b = tspan
    h = (b - a) / n
    t = np.linspace(a, b, n + 1)

    # Initialize output.
    u = np.tile(np.array(u0), (n+1, 1))

    # Time stepping.
    for i in range(n):
        k1 = h * du_dt(t[i], u[i])
        k2 = h * du_dt(t[i] + h / 2, u[i] + k1 / 2)
        k3 = h * du_dt(t[i] + h / 2, u[i] + k2 / 2)
        k4 = h * du_dt(t[i] + h, u[i] + k3)
        u[i+1] = u[i] + (k1 + 2 * (k2 + k3) + k4) / 6

    return t, u.T

def rk23(du_dt, tspan, u0, tol):
    """
    rk23(du_dt, tspan, u0, tol)

    Apply adaptive embedded RK formula to solve the vector-valued IVP u'=du_dt(u,p,t)
    over the interval tspan with u(tspan[1])=u0, with error tolerance tol.
    Return a vector of times and a vector of solution values/vectors.
    """
    # Initialize for the first time step.
    t = [tspan[0]]
    u = np.array([u0])
    i = 0
    h = 0.5 * tol ** (1 / 3)
    s1 = du_dt(t[0], u[0])

    # Time stepping.
    while t[i] < tspan[-1]:
        # Detect underflow of the step size.
        if t[i] + h == t[i]:
            warnings.warn(f"Stepsize too small near t={t[i]}")
            break  # quit time stepping loop

        # New RK stages.
        s2 = du_dt(t[i] + h / 2, u[i] + (h / 2) * s1)
        s3 = du_dt(t[i] + 3 * h / 4, u[i] + (3 * h / 4) * s2)
        unew2 = u[i] + h * (2 * s1 + 3 * s2 + 4 * s3) / 9  # 2rd order solution
        s4 = du_dt(t[i] + h, unew2)
        err = h * (-5 * s1 / 72 + s2 / 12 + s3 / 9 - s4 / 8)  # 2nd/3rd order difference
        E = np.linalg.norm(err, np.inf)  # error estimate
        maxerr = tol * (1 + np.linalg.norm(u[i], np.inf))  # relative/absolute blend

        # Accept the proposed step?
        if E < maxerr:  # yes
            t.append(t[i] + h)
            u = np.vstack([u, unew2])
            i += 1
            s1 = s4  # use FSAL property

        # Adjust step size.
        q = 0.8 * (maxerr / E) ** (1 / 3)  # conservative optimal step factor
        q = min(q, 4)  # limit stepsize growth
        h = min(q * h, tspan[-1] - t[i])  # don't step past the end

    # Convert outputs to arrays
    return np.array(t), u.T

def ab4(du_dt, tspan, u0, n):
    """
    ab4(du_dt, tspan, u0, n)

    Apply the Adams-Bashforth 4th order method to solve the vector-valued IVP u'=du_dt(u,p,t)
    over the interval tspan with u(tspan[1])=u0, using n subintervals/steps.
    """
    # Time discretization.
    a, b = tspan
    h = (b - a) / n
    t = np.linspace(a, b, n + 1)

    # Constants in the AB4 method.
    k = 4
    sigma = np.array([55, -59, 37, -9]) / 24

    # Find starting values by RK4.
    ts, us = rk4(du_dt, [a, a + (k - 1) * h], u0, k - 1)
    u = np.tile(np.array(u0), (n+1, 1))
    u[:k] = us[:k].T

    # Compute history of u' values, from newest to oldest.
    f = np.array([du_dt(t[k-j-2], u[k-j-2]) for j in range(k)])

    # Time stepping.
    for i in range(k-1, n):
        f = np.vstack([du_dt(t[i], u[i]), f[:-1]])  # new value of du/dt
        u[i+1] = u[i] + h * np.dot(sigma, f)  # advance one step

    return t, u.T

def am2(du_dt, tspan, u0, n):
    """
    am2(du_dt, tspan, u0, n)

    Apply the Adams-Moulton 2nd order method to solve the vector-valued IVP u'=du_dt(u,p,t)
    over the interval tspan with u(tspan[1])=u0, using n subintervals/steps.
    """
    # Time discretization.
    a, b = tspan
    h = (b - a) / n
    t = np.linspace(a, b, n + 1)

    # Initialize output.
    u = np.tile(np.array(u0), (n+1, 1))

    # Time stepping.
    for i in range(n):
        # Data that does not depend on the new value.
        known = u[i] + h / 2 * du_dt(t[i], u[i])
        # Find a root for the new value.
        F = lambda z: z - h / 2 * du_dt(t[i+1], z) - known
        unew = levenberg(F, known)
        u[i+1] = unew[-1]

    return t, u.T



import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np
from numpy.linalg import norm

def test_poissonfd():
    f = lambda x,y: -np.sin(3*x*y-4*y)*(9*y**2+(3*x-4)**2)
    g = lambda x,y: np.sin(3*x*y-4*y)
    xspan = [0,1]
    yspan = [0,2]
    U, X, Y = FNC.poissonfd(f,g,100,xspan,90,yspan)
    assert norm(g(X,Y) - U) / norm(U) < 1e-3

def test_elliptic():
    lamb = 1.5
    def pde(X,Y,U,Ux,Uxx,Uy,Uyy):
        return Uxx + Uyy - lamb/(U+1)**2   # residual
    g = lambda x,y: x + y     # boundary condition
    u = FNC.elliptic(pde, g, 30, [0,2.5], 24, [0,1])
    assert isclose(u(1.25, 0.5), 1.7236921361, rtol=1e-6)
    assert isclose(u(1, 0), 1, rtol=1e-6)



from numpy.linalg import norm, lstsq
import numpy as np
import warnings

def newton(f, dfdx, x1):
    """
    newton(f, dfdx, x1)

    Use Newton's method to find a root of f starting from x1, where dfdx is the
    derivative of f. Returns a vector of root estimates.
    """
    # Operating parameters.
    eps = np.finfo(float).eps
    funtol = 100 * eps
    xtol = 100 * eps
    maxiter = 40

    x = np.zeros(maxiter)
    x[0] = x1
    y = f(x1)
    dx = np.inf  # for initial pass below
    k = 0

    while (abs(dx) > xtol) and (abs(y) > funtol) and (k < maxiter):
        dydx = dfdx(x[k])
        dx = -y / dydx  # Newton step
        x[k + 1] = x[k] + dx  # new estimate

        k = k + 1
        y = f(x[k])

    if k == maxiter:
        warnings.warn("Maximum number of iterations reached.")

    return x[:k+1]

def secant(f, x1, x2):
    """
    secant(f, x1, x2)

    Use the secant method to find a root of f starting from x1 and x2. Returns a
    vector of root estimates.
    """
    # Operating parameters.
    eps = np.finfo(float).eps
    funtol = 100 * eps
    xtol = 100 * eps
    maxiter = 40

    x = np.zeros(maxiter)
    x[:2] = [x1, x2]
    y1 = f(x1)
    y2 = 100
    dx = np.inf  # for initial pass below
    k = 1

    while (abs(dx) > xtol) and (abs(y2) > funtol) and (k < maxiter):
        y2 = f(x[k])
        dx = -y2 * (x[k] - x[k - 1]) / (y2 - y1)  # secant step
        x[k + 1] = x[k] + dx  # new estimate

        k = k + 1
        y1 = y2  # current f-value becomes the old one next time

    if k == maxiter:
        warnings.warn("Maximum number of iterations reached.")
    return x[:k+1]

def newtonsys(f, jac, x1):
    """
        newtonsys(f, jac, x1)

    Use Newton's method to find a root of a system of equations, starting from x1. The
    function f should return the residual vector, and the function jac should return 
    the Jacobian matrix. Returns root estimates as a matrix, one estimate per column.
    """
    # Operating parameters.
    funtol = 1000 * np.finfo(float).eps
    xtol = 1000 * np.finfo(float).eps
    maxiter = 40

    x = np.zeros((maxiter, len(x1)))
    x[0] = x1
    y, J = f(x1), jac(x1)
    dx = 10.0  # for initial pass below
    k = 0

    while (norm(dx) > xtol) and (norm(y) > funtol) and (k < maxiter):
        dx = -lstsq(J, y)[0]  # Newton step
        x[k+1] = x[k] + dx

        k = k + 1
        y, J = f(x[k]), jac(x[k])

    if k == maxiter:
        warnings.warn("Maximum number of iterations reached.")
    return x[:k+1]

def fdjac(f, x0, y0):
    """
    fdjac(f,x0,y0)

    Compute a finite-difference approximation of the Jacobian matrix for f at x0,
    where y0=f(x0) is given.
    """

    delta = np.sqrt(np.finfo(float).eps)  # FD step size
    m, n = len(y0), len(x0)
    J = np.zeros((m, n))
    I = np.eye(n)
    for j in range(n):
        J[:, j] = (f(x0 + delta * I[:, j]) - y0) / delta
    return J

def levenberg(f, x1, tol=1e-12):
    """
    levenberg(f,x1,tol)

    Use Levenberg's quasi-Newton iteration to find a root of the system f, 
    starting from x1, with tol as the stopping tolerance in both step size and residual norm. Returns root estimates as a matrix, one estimate per column.
    """

    # Operating parameters.
    ftol = tol
    xtol = tol
    maxiter = 40

    n = len(x1)
    x = np.zeros((maxiter+1, n))
    x[0] = x1
    fk = f(x1)
    k = 0
    s = 10.0
    Ak = fdjac(f, x[0], fk)  # start with FD Jacobian
    jac_is_new = True

    lam = 10
    while (norm(s) > xtol) and (norm(fk) > ftol) and (k < maxiter):
        # Compute the proposed step.
        B = Ak.T @ Ak + lam * np.eye(n)
        z = Ak.T @ fk
        s = -lstsq(B, z)[0]

        xnew = x[k] + s
        fnew = f(xnew)

        # Do we accept the result?
        if norm(fnew) < norm(fk):  # accept
            y = fnew - fk
            x[k + 1] = xnew
            fk = fnew
            k = k + 1

            lam = lam / 10  # get closer to Newton
            # Broyden update of the Jacobian.
            Ak = Ak + np.outer(y - Ak @ s, s / np.dot(s, s))
            jac_is_new = False
        else:  # don't accept
            # Get closer to steepest descent.
            lam = lam * 4
            # Re-initialize the Jacobian if it's out of date.
            if not jac_is_new:
                Ak = fdjac(f, x[k], fk)
                jac_is_new = True

    if norm(fk) > 1e-3:
        warnings.warn("Iteration did not find a root.")
    return x[:k+1]



import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np

A = np.array([[ 1.0, 2, 3, 0], [-1, 1, 2, -1], [3, 1, 2, 4], [1, 1, 1, 1] ])

def test_lufact():
    L, U = FNC.lufact(A)
    assert isclose(L @ U, A).all()
    assert isclose(np.tril(L), L).all()
    assert isclose(np.triu(U), U).all()

def test_trisub():
    L = np.array([[1.0, 0, 0, 0], [-1, 1, 0, 0], [3, 1, 1, 0], [1, 1, 1, 1]])
    x = np.array([1, -2.5, 0, 2])
    result = FNC.forwardsub(L, L @ x)
    assert isclose(result, x).all()
    U = np.array([[1.0, 2, 3, 0], [0, 2, 2, -1], [0, 0, -1, 4], [0, 0, 0, 1]])
    result = FNC.backsub(U, U @ x)
    assert isclose(result, x).all()

def test_plufact():
    L, U, p = FNC.plufact(A)
    assert isclose(L @ U, A[p, :]).all()
    assert isclose(np.tril(L), L).all()
    assert isclose(np.triu(U), U).all()




import scipy
import numpy as np
from .chapter02 import forwardsub, backsub

def lsnormal(A, b):
    """
    lsnormal(A, b)
    
    Solve a linear least squares problem by the normal equations. Returns the
    minimizer of ||b-Ax||.
    """
    N = A.T @ A
    z = A.T @ b
    R = scipy.linalg.cholesky(N)
    w = forwardsub(R.T, z)                   # solve R'z=c
    x = backsub(R, w)                        # solve Rx=z
    return x

def lsqrfact(A, b):
    """
    lsqrfact(A, b)
    
    Solve a linear least squares problem by QR factorization. Returns the
    minimizer of ||b-Ax||.
    """
    Q, R = np.linalg.qr(A)
    c = Q.T @ b
    x = backsub(R, c)
    return x

def qrfact(A):
    """
        qrfact(A)

    QR factorization by Householder reflections. Returns Q and R.
    """
    m, n = A.shape
    Qt = np.eye(m)
    R = np.copy(A)
    for k in range(n):
        z = R[k:, k]
        w = np.hstack((-np.sign(z[0]) * np.linalg.norm(z) - z[0], -z[1:]))
        nrmw = np.linalg.norm(w)
        if nrmw < np.finfo(float).eps: continue    # skip this iteration
        v = w / nrmw
        # Apply the reflection to each relevant column of R and Q
        for j in range(k, n):
            R[k:, j] -= 2 * np.dot(v, R[k:, j]) * v
        for j in range(m):
            Qt[k:, j] -= 2 * np.dot(v, Qt[k:, j]) * v 
    return Qt.T, np.triu(R)



import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np

def test_polyinterp():
    f = lambda x: np.exp(np.sin(x)+x**2)
    t = np.array([-np.cos(k*np.pi/40) for k in range(41)])
    p = FNC.polyinterp(t,f(t))
    assert isclose(p(-0.12345), f(-0.12345))

def test_triginterp():
    f = lambda x: np.exp(np.sin(np.pi*x))
    n = 30
    t = np.array([ 2*k/(2*n+1) for k in range(-n,n+1) ])
    p = FNC.triginterp(t,f(t))
    assert isclose(p(-0.12345), f(-0.12345))
    t = np.array([ k/n for k in range(-n,n) ])
    p = FNC.triginterp(t,f(t))
    assert isclose(p(-0.12345), f(-0.12345))

def test_ccint():
    F = lambda x: np.tan(x/2-0.2)
    f = lambda x: 0.5/np.cos(x/2-0.2)**2
    assert isclose(FNC.ccint(f,40)[0], F(1)-F(-1))

def test_glint():
    F = lambda x: np.tan(x/2-0.2)
    f = lambda x: 0.5/np.cos(x/2-0.2)**2
    assert isclose(FNC.glint(f,40)[0], F(1)-F(-1))

def test_intinf():
    f = lambda x: 1/(32+2*x**4)
    assert isclose(FNC.intinf(f,1e-9)[0], np.sqrt(2)*np.pi/32, rtol=1e-5)

def test_intsing():
    f = lambda x: (1-x)/( np.sin(x)**0.5 )
    assert isclose(FNC.intsing(f,1e-8)[0], 1.34312, rtol=1e-5)



import numpy as np
import scipy
import scipy.sparse as sp
from .chapter10 import diffmat2, diffcheb
from .chapter04 import levenberg


def tensorgrid(x, y):
    """
    tensorgrid(x, y)

    Create a tensor grid for a rectangle from its 1d projections x and y.
    Returns a function to reshape a 2d array to a vector, a function to reshape 
    a vector into a 2d array, a function to evaluate a function on the grid, 
    two arrays to give the grid coordinates, and a boolean array to identify 
    the boundary points.
    """
    m, n = len(x) - 1, len(y) - 1
    vec = lambda U: U.T.flatten()
    unvec = lambda u: np.reshape(u, (n+1, m+1)).T
    mtx = lambda h: np.array([[h(xi ,yj) for yj in y] for xi in x])
    X = mtx(lambda x, y: x)
    Y = mtx(lambda x, y: y)

    # Identify boundary points.
    is_boundary = np.tile(True, (m+1, n+1))
    is_boundary[1:-1, 1:-1] = False
    
    return mtx, X, Y, vec, unvec, is_boundary


def poissonfd(f, g, m, xspan, n, yspan):
    """
    poissonfd(f, g, m, xspan, n, yspan)

    Solve Poisson's equation on a rectangle by finite differences. Function f is the
    forcing function and function g gives the  Dirichlet boundary condition. The rectangle
    is the tensor product of intervals xspan and yspan,  and the discretization uses
    m+1 and n+1 points in the two coordinates.

    Return matrices of the solution values, and the coordinate functions, on the grid.
    """
    # Discretize the domain.
    x, Dx, Dxx = diffmat2(m, xspan)
    y, Dy, Dyy = diffmat2(n, yspan)
    mtx, X, Y, vec, unvec, is_boundary = tensorgrid(x, y)
    N = (m+1) * (n+1)    # total number of unknowns

    # Form the collocated PDE as a linear system.
    Dxx = sp.lil_matrix(Dxx)
    Dyy = sp.lil_matrix(Dyy)
    A = sp.kron(sp.eye(n+1, format="lil"), Dxx) + sp.kron(Dyy, sp.eye(m+1, format="lil"))
    b = vec(mtx(f))

    # Apply Dirichlet condition.
    idx = vec(is_boundary)
    scale = np.max(abs(A[n+1, :]))
    I = sp.eye(N, format="lil")
    A[idx, :] = scale * I[idx, :]         # Dirichet assignment
    X_bd, Y_bd = vec(X)[idx], vec(Y)[idx]
    b[idx] = scale * g(X_bd, Y_bd)    # assigned values

    # Solve the linear sytem and reshape the output.
    u = scipy.sparse.linalg.spsolve(A.tocsr(), b)
    U = unvec(u)
    return U, X, Y


def elliptic(f, g, m, xspan, n, yspan):
    """
    newtonpde(f, g, m, xspan, n, yspan)

    Newton's method with finite differences to solve the PDE f(u,x,y,disc)=0 on the
    rectangle xspan \times yspan, subject to g(x,y)=0 on the boundary. Use m+1
    points in x by n+1 points in y.

    Return matrices of the solution values, and the coordinate functions, on the grid.
    """
    from scipy.sparse.linalg import spsolve
    x, Dx, Dxx = diffcheb(m, xspan)
    y, Dy, Dyy = diffcheb(n, yspan)
    mtx, X, Y, vec, unvec, is_boundary = tensorgrid(x, y)

    # Evaluate the boundary condition at the boundary nodes.
    idx = vec(is_boundary)
    X_bd, Y_bd = vec(X)[idx], vec(Y)[idx]
    g_bd = g(X_bd, Y_bd)

    # Evaluate the PDE+BC residual.
    def residual(u):
        U = unvec(u)
        R = f(X, Y, U, Dx @ U, Dxx @ U, U @ Dy.T, U @ Dyy.T)    # PDE
        r = vec(R)
        r[idx] = u[idx] - g_bd                                  # BC
        return r
    
    # Solve the equation.
    u = levenberg(residual, vec(np.zeros(X.shape)))[-1]
    U = unvec(u)

    def evaluate(xi, eta):
        v = [chebinterp(y, u, eta) for u in U]
        return chebinterp(x, v, xi)
    
    return np.vectorize(evaluate)

def chebinterp(x, v, xi):
    "Evaluate Chebyshev interpolant with nodes x, values v, at point xi"
    n = len(x) - 1
    w = np.ones(n+1)
    w[1::2] = -1    # alternating ±1
    w[[0, n]] *= 0.5
    if xi in x:    # exactly at a node
        return v[np.where(x == xi)[0][0]]
    else:
        terms = w / (xi - x)
        return np.sum(v * terms) / np.sum(terms)



import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np
from scipy.interpolate import interp1d

f = lambda t: np.cos(5*t)
t = np.array([-2,-0.5,0,1,1.5,3.5,4]) / 10

def test_intadapt():
    Q, t = FNC.intadapt(f, -1, 3, 1e-8)
    assert isclose(Q, (np.sin(15) + np.sin(5)) / 5, rtol=1e-5)

def test_trapezoid():
    T, _, _ = FNC.trapezoid(f, -1, 3, 820)
    assert isclose(T, (np.sin(15) + np.sin(5)) / 5, rtol=1e-4)

def test_fdweights():
    w = FNC.fdweights(t - 0.12, 2)
    g = lambda t: np.cos(3*t)
    val = np.dot(w, g(t))
    assert isclose(val, -9*np.cos(0.36), rtol=1e-3)

def test_hatfun():
    H = FNC.hatfun(t, 5)
    P = interp1d(t, [0,0,0,0,0,1,0], kind='linear')
    x = np.array([0.22, 0.38, 0.06, t[5], t[6], t[0]])
    assert isclose(H(x), P(x)).all()

def test_plinterp():
    p = FNC.plinterp(t, f(t))
    P = interp1d(t, f(t), kind='linear')
    x = [-0.2, 0, 0.22, 0.38]
    assert isclose(p(x), P(x)).all()

def test_spinterp():
    S = FNC.spinterp(t, np.exp(t))
    x = np.array([-.17, -0.01, 0.33, .38])
    assert isclose(S(x), np.exp(x), rtol=1e-5).all()
    assert isclose(S(t), np.exp(t), rtol=1e-11).all()



import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np

f = lambda x: np.exp(x**2 - 3*x)
df = lambda x: (2*x - 3) * f(x)
ddf = lambda x: ((2*x - 3)**2 + 2) * f(x)

def test_shoot():
    lamb = 0.6
    phi = lambda r, w, dwdr: lamb / w**2 - dwdr / r
    a, b = 1e-15, 1
    g1 = lambda u, du: du
    g2 = lambda u, du: u - 1
    r, w, dwdr = FNC.shoot(phi, a, b, g1, g2, [0.8, 0])
    assert isclose(w[0], 0.78776, rtol=1e-4)

def test_diffmat2():
    x, D, DD = FNC.diffmat2(400, (-0.5, 2))
    assert isclose(df(x), D @ f(x), rtol=1e-3).all()
    assert isclose(ddf(x), DD @ f(x), rtol=1e-3).all()

def test_diffcheb():
    t, D, DD = FNC.diffcheb(80, (-0.5, 2))
    assert isclose(df(t), D @ f(t), rtol=1e-7).all()
    assert isclose(ddf(t), DD @ f(t), rtol=1e-7).all()

def test_bvplin():
    exact = lambda x: np.exp(np.sin(x))
    p = lambda x: -np.cos(x)
    q = np.sin
    r = lambda x: np.zeros(x.shape)
    x, u = FNC.bvplin(p, q, r, [0, np.pi/2], 1, np.exp(1), 300)
    assert isclose(u, exact(x), rtol=1e-3).all()

def test_bvp():
    phi = lambda t, theta, omega: -0.05 * omega - np.sin(theta)
    g1 = lambda u, du: u - 2.5
    g2 = lambda u, du: u + 2
    init = np.linspace(2.5, -2, 101)
    t, theta = FNC.bvp(phi, [0, 5], g1, g2, init)
    assert isclose(theta[6], 2.421850016880724, rtol=1e-5)

def test_fem():
    c = lambda x: x**2
    q = lambda x: np.tile(4, x.shape)
    f = lambda x: np.sin(np.pi * x)
    x, u = FNC.fem(c, q, f, 0, 1, 100)
    assert isclose(u[32], 0.1641366907307196, rtol=1e-10)




import warnings
import numpy as np
from numpy.linalg import eig
from .chapter05 import intadapt


def polyinterp(t, y):
    """
    polyinterp(t, y)

    Return a callable polynomial interpolant through the points in vectors t, y. Uses
    the barycentric interpolation formula.
    """
    n = len(t) - 1
    C = (t[-1] - t[0]) / 4  # scaling factor to ensure stability
    tc = t / C

    # Adding one node at a time, compute inverses of the weights.
    omega = np.ones(n + 1)
    for m in range(n):
        d = tc[: m + 1] - tc[m + 1]  # vector of node differences
        omega[: m + 1] = omega[: m + 1] * d  # update previous
        omega[m + 1] = np.prod(-d)  # compute the new one
    w = 1 / omega  # go from inverses to weights

    def p(x):
        # Compute interpolant.
        z = np.where(x == t)[0]
        if len(z) > 0:  # avoid dividing by zero
            # Apply L'Hopital's Rule exactly.
            f = y[z[0]]
        else:
            terms = w / (x - t)
            f = np.sum(y * terms) / np.sum(terms)
        return f

    return np.vectorize(p)


def triginterp(t, y):
    """
        triginterp(t, y)

    Return trigonometric interpolant for points defined by vectors t and y.
    """
    N = len(t)

    def trigcardinal(x):
        if x == 0:
            tau = 1.0
        elif np.mod(N, 2) == 1:  # odd
            tau = np.sin(N * np.pi * x / 2) / (N * np.sin(np.pi * x / 2))
        else:  # even
            tau = np.sin(N * np.pi * x / 2) / (N * np.tan(np.pi * x / 2))
        return tau

    def p(x):
        return np.sum([y[k] * trigcardinal(x - t[k]) for k in range(N)])

    return np.vectorize(p)


def ccint(f, n):
    """
    ccint(f, n)

    Perform Clenshaw-Curtis integration for the function f on n+1 nodes in [-1,1]. Return
    integral and a vector of the nodes used. Note: n must be even.
    """
    # Find Chebyshev extreme nodes.
    theta = np.linspace(0, np.pi, n + 1)
    x = -np.cos(theta)

    # Compute the C-C weights.
    c = np.zeros(n + 1)
    c[[0, n]] = 1 / (n**2 - 1)
    v = np.ones(n - 1)
    for k in range(1, int(n / 2)):
        v -= 2 * np.cos(2 * k * theta[1:-1]) / (4 * k**2 - 1)
    v -= np.cos(n * theta[1:-1]) / (n**2 - 1)
    c[1:-1] = 2 * v / n

    # Evaluate integrand and integral.
    I = np.dot(c, f(x))  # use vector inner product
    return I, x


def glint(f, n):
    """
    glint(f, n)

    Perform Gauss-Legendre integration for the function f on n nodes in (-1,1). Return
    integral and a vector of the nodes used.
    """
    # Nodes and weights are found via a tridiagonal eigenvalue problem.
    beta = 0.5 / np.sqrt(1 - (2.0 * np.arange(1, n)) ** (-2))
    T = np.diag(beta, 1) + np.diag(beta, -1)
    ev, V = eig(T)
    ev = np.real_if_close(ev)
    p = np.argsort(ev)
    x = ev[p]  # nodes
    c = 2 * V[0, p] ** 2  # weights

    # Evaluate the integrand and compute the integral.
    I = np.dot(c, f(x))  # vector inner product
    return I, x

def intinf(f, tol):
    """
    intinf(f, tol)

    Perform doubly-exponential integration of function f over (-Inf,Inf), using
    error tolerance tol. Return integral and a vector of the nodes used.
    """
    xi = lambda t: np.sinh(np.sinh(t))
    dxi_dt = lambda t: np.cosh(t) * np.cosh(np.sinh(t))
    g = lambda t: f(xi(t)) * dxi_dt(t)
    M = 3
    while (abs(g(-M)) > tol/100) or (abs(g(M)) > tol/100):
        M += 0.5
        if np.isinf(xi(M)):
            warnings.warn("Function may not decay fast enough.")
            M -= 0.5
            break

    I, t = intadapt(g,-M,M,tol)
    x = xi(t)
    return I, x

def intsing(f, tol):
    """
    intsing(f, tol)

    Adaptively integrate function f over (0,1), where f may be 
    singular at zero, with error tolerance tol. Returns the
    integral estimate and a vector of the nodes used.
    """
    xi = lambda t: 2 / (1 + np.exp( 2*np.sinh(t) ))
    dxi_dt = lambda t: np.cosh(t) / np.cosh(np.sinh(t))**2
    g = lambda t: f(xi(t)) * dxi_dt(t)
    # Find where to truncate the integration interval.
    M = 3
    while abs(g(M)) > tol/100:
        M += 0.5
        if xi(M) == 0:
            warnings.warn("Function may grow too rapidly.")
            M -= 0.5
            break

    I, t = intadapt(g, 0, M, tol)
    x = xi(t)
    return I, x



import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np

A = np.array([[3.0, 4, 5], [-1, 0, 1], [4, 2, 0], [1, 1, 2], [3, -4, 1] ])
b = np.array([5.0, 4, 3, 2, 1])
x_true = np.linalg.lstsq(A, b)[0]

def test_lsnormal():
    x = FNC.lsnormal(A, b)
    assert isclose(x_true, x, rtol=1e-6).all()

def test_lsqrfact():
    x = FNC.lsqrfact(A, b)
    assert isclose(x_true, x).all()
    
def test_qrfact():
    Q, R = FNC.qrfact(A)
    assert isclose(Q @ R, A).all()
    assert isclose(Q.T @ Q, np.eye(Q.shape[0])).all()
    assert isclose(R, np.triu(R)).all()
    assert R.shape == (5, 3)



import numpy as np
# from numpy.linalg import norm, solve, lstsq
from scipy.linalg import lu
from scipy.sparse import csc_matrix, diags


def poweriter(A, numiter):
    """
    poweriter(A, numiter)

    Perform numiter power iterations with the matrix A, starting from a random vector, 
    and return a vector of eigenvalue estimates and the final eigenvector approximation.
    """
    n = A.shape[0]
    x = np.random.randn(n)
    x = x / np.linalg.norm(x)
    beta = np.zeros(numiter)
    for k in range(numiter):
        y = A @ x
        beta[k] = np.dot(x, y)
        x = y / np.linalg.norm(y)

    return beta, x


def inviter(A, s, numiter):
    """
    inviter(A, s, numiter)

    Perform numiter inverse iterations with the matrix A and shift s, starting
    from a random vector, and return a vector of eigenvalue estimates and the final
    eigenvector approximation.
    """
    n = A.shape[0]
    x = np.random.randn(n)
    x = x / np.linalg.norm(x)
    beta = np.zeros(numiter)
    PL, U = lu(A - s * np.eye(n), permute_l=True)
    for k in range(numiter):
        y = np.linalg.solve(U, np.linalg.solve(PL, x))
        alpha = np.dot(x, y)
        beta[k] = (1 / alpha) + s
        x = y / np.linalg.norm(y)

    return beta, x


def arnoldi(A, u, m):
    """
    arnoldi(A, u, m)

    Perform the Arnoldi iteration for A starting with vector u, out to the Krylov
    subspace of degree m. Return the orthonormal basis (m+1 columns) and the upper
    Hessenberg H of size m+1 by m.
    """
    n = u.size
    Q = np.zeros([n, m + 1])
    H = np.zeros([m + 1, m])
    Q[:, 0] = u / np.linalg.norm(u)
    for j in range(m):
        # Find the new direction that extends the Krylov subspace.
        v = A @ Q[:, j]
        # Remove the projections onto the previous vectors.
        for i in range(j + 1):
            H[i, j] = Q[:, i] @ v
            v -= H[i, j] * Q[:, i]
        # Normalize and store the new basis vector.
        H[j + 1, j] = np.linalg.norm(v)
        Q[:, j + 1] = v / H[j + 1, j]

    return Q, H


def gmres(A, b, m):
    """
    gmres(A, b, m)

    Do m iterations of GMRES for the linear system A*x=b. Return the final solution
    estimate x and a vector with the history of residual norms. (This function is for
    demo only, not practical use.)
    """
    n = len(b)
    Q = np.zeros([n, m + 1])
    Q[:, 0] = b / np.linalg.norm(b)
    H = np.zeros([m + 1, m])

    # Initial "solution" is zero.
    residual = np.hstack([np.linalg.norm(b), np.zeros(m)])

    for j in range(m):
        # Next step of Arnoldi iteration.
        v = A @ Q[:, j]
        for i in range(j + 1):
            H[i, j] = Q[:, i] @ v
            v -= H[i, j] * Q[:, i]
        H[j + 1, j] = np.linalg.norm(v)
        Q[:, j + 1] = v / H[j + 1, j]

        # Solve the minimum residual problem.
        r = np.hstack([np.linalg.norm(b), np.zeros(j + 1)])
        z = np.linalg.lstsq(H[:j + 2, :j + 1], r)[0]
        x = Q[:, :j + 1] @ z
        residual[j + 1] = np.linalg.norm(A @ x - b)

    return x, residual


def sprandsym(n, density, **kwargs):
    if "rcond" in kwargs:
        ev = np.array([kwargs["rcond"] ** (i / (n - 1)) for i in range(n)])
    elif "eigvals" in kwargs:
        ev = kwargs["eigvals"]
    else:
        ev = np.sqrt(n) * np.random.randn(n)

    def randjr(A):
        # Random Jacobi rotation similarity transformation.
        theta = 2 * np.pi * np.random.rand()
        c = np.cos(theta)
        s = np.sin(theta)
        idx = [int(k) for k in np.floor(n * np.random.rand(2))]
        R = np.array([[c, s], [-s, c]])
        A[idx, :] = R @ A[idx, :]
        A[:, idx] = A[:, idx] @ R.T
        return A

    targetnz = np.ceil(min(0.98, density) * n * n)
    A = diags(ev, 0, format="lil")
    while A.nnz < targetnz:
        A = randjr(A)

    return csc_matrix(A)

def poisson2d(m):
    n = m**2
    d0 = [4.0] * n
    d1 = [-1.0] * (n-1)
    d1[m-1::m] = np.zeros(m-1)
    dm = np.array([-1.0] * (n - m))
    return (m + 1)**2 / np.pi**2 * diags([d0, d1, d1, dm, dm], [0, 1, -1, m, -m], format="csc")




import numpy as np

def forwardsub(L,b):
    """
     forwardsub(L,b)

    Solve the lower-triangular linear system with matrix L and right-hand side
    vector b.
    """
    n = len(b)
    x = np.zeros(n)
    for i in range(n):
        s = L[i,:i] @ x[:i]
        x[i] = ( b[i] - s ) / L[i, i]
    return x


def backsub(U,b):
    """
    backsub(U,b)

    Solve the upper-triangular linear system with matrix U and right-hand side
    vector b.
    """
    n = len(b)
    x = np.zeros(n)
    for i in range(n-1, -1, -1):
        s = U[i, i+1:] @ x[i+1:]
        x[i] = ( b[i] - s ) / U[i, i]
    return x

def lufact(A):
    """
    lufact(A)

    Compute the LU factorization of square matrix A, returning the
    factors.
    """
    n = A.shape[0]     # detect the dimensions from the input
    L = np.eye(n)      # ones on main diagonal, zeros elsewhere
    U = np.zeros((n, n))
    A_k = np.copy(A.astype(float))    # make a working copy 

    # Reduction by np.outer products
    for k in range(n-1):
        U[k, :] = A_k[k, :]
        L[:, k] = A_k[:, k] / U[k,k]
        A_k -= np.outer(L[:,k], U[k,:])
    U[n-1, n-1] = A_k[n-1, n-1]
    return L, U

def plufact(A):
    """
        plufact(A)

    Compute the PLU factorization of square matrix A, returning the
    triangular factors and a row permutation vector.
    """
    n = A.shape[0]
    L = np.zeros((n, n))
    U = np.zeros((n, n))
    p = np.zeros(n, dtype=int)
    A_k = np.copy(A.astype(float))

    # Reduction by outer products
    for k in range(n):
        p[k] = np.argmax(abs(A_k[:, k]))
        U[k, :] = A_k[p[k], :]
        L[:, k] = A_k[:, k] / U[k, k]
        if k < n-1:
            A_k -= np.outer(L[:, k], U[k, :])
    return L[p, :], U, p


def horner(c,x):
    """
    horner(c,x)

    Evaluate a polynomial whose coefficients are given in descending order 
    in c, at the point x, using Horner's rule.
    """
    n = len(c)
    y = c[0]
    for k in range(1, n):
        y = x * y + c[k]   
    return y


from .chapter01 import *
from .chapter02 import *
from .chapter03 import *
from .chapter04 import *
from .chapter05 import *
from .chapter06 import *
from .chapter08 import *
from .chapter09 import *
from .chapter10 import *
from .chapter11 import *
from .chapter13 import *



import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np

def test_diffper():
    s = lambda x: np.sin(np.pi*(x-0.2))
    c = lambda x: np.cos(np.pi*(x-0.2))
    f = lambda x: 1 + s(x)**2
    df = lambda x: 2*np.pi*s(x)*c(x)
    ddf = lambda x: 2*np.pi**2*(c(x)**2 - s(x)**2)
    t,D,DD = FNC.diffper(400,(0,2))
    assert isclose(df(t), D @ f(t), rtol=1e-3).all()
    assert isclose(ddf(t), DD @ f(t), rtol=1e-3).all()

def test_parabolic():
    phi = lambda t,x,u,ux,uxx: uxx + t*u
    g1 = lambda u,ux: ux
    g2 = lambda u,ux: u-1
    init = lambda x: x**2
    x,u = FNC.parabolic(phi,(0,1),40,g1,g2,(0,2),init)
    assert isclose(u(0.5)[20], 0.845404, rtol=1e-3)
    assert isclose(u(1)[-1], 1, rtol=1e-4)
    assert isclose(u(2)[0], 2.45692, rtol=1e-3)



import pytest
import fncbook as FNC  # The code to test
from numpy import isclose
import numpy as np


def test_newton():
    for c in [2,4,7.5,11]:
        f = lambda x: np.exp(x) - x - c
        dfdx = lambda x: np.exp(x) - 1
        x = FNC.newton(f,dfdx,1.0)
        r = x[-1]
        assert isclose(f(r), 0)

def test_secant():
     for c in [2,4,7.5,11]:
        f = lambda x: np.exp(x) - x - c
        x = FNC.secant(f, 3, 0.5)
        r = x[-1]
        assert isclose(f(r), 0)

def test_systems():
    def nlfun(x):
        f = np.array([     
            np.exp(x[1]-x[0]) - 2,
            x[0]*x[1] + x[2],
            x[1]*x[2] + x[0]**2 - x[1]
        ])
        return f
    def nljac(x):
        J = np.zeros((3,3))
        J[0] = [-np.exp(x[1]-x[0]), np.exp(x[1]-x[0]), 0]
        J[1] = [x[1], x[0], 1]
        J[2] = [2*x[0], x[2]-1, x[1]]
        return J
    x = FNC.newtonsys(nlfun,nljac,np.array([0.,0,0]))
    assert isclose(nlfun(x[-1]), [0, 0, 0]).all()
    x = FNC.newtonsys(nlfun,nljac,np.array([1.,2,3]))
    assert isclose(nlfun(x[-1]), [0, 0, 0]).all()
    x = FNC.levenberg(nlfun, np.array([10.,-4,-3]))
    assert isclose(nlfun(x[-1]), [0, 0, 0]).all()
